package jalview.datamodel.features;

import jalview.datamodel.SequenceFeature;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * A data store for a set of sequence features that supports efficient lookup of
 * features overlapping a given range. Intended for (but not limited to) storage
 * of features for one sequence and feature type.
 * 
 * @author gmcarstairs
 *
 */
public class FeatureStore
{
  Comparator<ContiguousI> startOrdering = new RangeComparator(true);

  Comparator<ContiguousI> endOrdering = new RangeComparator(false);

  /*
   * Non-positional features have no (zero) start/end position.
   * Kept as a separate list in case this criterion changes in future.
   */
  List<SequenceFeature> nonPositionalFeatures;

  /*
   * An ordered list of features, with the promise that no feature in the list 
   * properly contains any other. This constraint allows bounded linear search
   * of the list for features overlapping a region.
   * Contact features are not included in this list.
   */
  List<SequenceFeature> nonNestedFeatures;

  /*
   * contact features ordered by first contact position
   */
  List<SequenceFeature> contactFeatureStarts;

  /*
   * contact features ordered by second contact position
   */
  List<SequenceFeature> contactFeatureEnds;

  /*
   * Nested Containment List is used to hold any features that are nested 
   * within (properly contained by) any other feature. This is a recursive tree
   * which supports depth-first scan for features overlapping a range.
   * It is used here as a 'catch-all' fallback for features that cannot be put
   * into a simple ordered list without invalidating the search methods.
   */
  NCList<SequenceFeature> nestedFeatures;

  /*
   * Feature groups represented in stored positional features 
   * (possibly including null)
   */
  Set<String> positionalFeatureGroups;

  /*
   * Feature groups represented in stored non-positional features 
   * (possibly including null)
   */
  Set<String> nonPositionalFeatureGroups;

  /**
   * Constructor
   */
  public FeatureStore()
  {
    nonNestedFeatures = new ArrayList<SequenceFeature>();
    positionalFeatureGroups = new HashSet<String>();

    // we only construct nonPositionalFeatures, contactFeatures
    // or the NCList if we need to
  }

  /**
   * Adds one sequence feature to the store, and returns true, unless the
   * feature is already contained in the store, in which case this method
   * returns false. Containment is determined by SequenceFeature.equals()
   * comparison.
   * 
   * @param feature
   */
  public boolean addFeature(SequenceFeature feature)
  {
    if (!feature.isNonPositional())
    {
      positionalFeatureGroups.add(feature.getFeatureGroup());
    }

    boolean added = false;

    if (feature.isContactFeature())
    {
      added = addContactFeature(feature);
    }
    else if (feature.isNonPositional())
    {
      added = addNonPositionalFeature(feature);
    }
    else
    {
      if (!nonNestedFeatures.contains(feature))
      {
        added = addNonNestedFeature(feature);
        if (!added)
        {
          /*
           * detected a nested feature - put it in the NCList structure
           */
          added = addNestedFeature(feature);
        }
      }
    }

    return added;
  }

  /**
   * Adds the feature to the list of non-positional features (with lazy
   * instantiation of the list if it is null), and returns true. If the
   * non-positional features already include the new feature (by equality test),
   * then it is not added, and this method returns false. The feature group is
   * added to the set of distinct feature groups for non-positional features.
   * 
   * @param feature
   */
  protected boolean addNonPositionalFeature(SequenceFeature feature)
  {
    if (nonPositionalFeatures == null)
    {
      nonPositionalFeatures = new ArrayList<SequenceFeature>();
      nonPositionalFeatureGroups = new HashSet<String>();
    }
    if (nonPositionalFeatures.contains(feature))
    {
      return false;
    }

    nonPositionalFeatures.add(feature);

    nonPositionalFeatureGroups.add(feature.getFeatureGroup());

    return true;
  }

  /**
   * Adds one feature to the NCList that can manage nested features (creating
   * the NCList if necessary), and returns true. If the feature is already
   * stored in the NCList (by equality test), then it is not added, and this
   * method returns false.
   */
  protected synchronized boolean addNestedFeature(SequenceFeature feature)
  {
    if (nestedFeatures == null)
    {
      nestedFeatures = new NCList<SequenceFeature>(feature);
      return true;
    }
    return nestedFeatures.add(feature, false);
  }

  /**
   * Add a feature to the list of non-nested features, maintaining the ordering
   * of the list. A check is made for whether the feature is nested in (properly
   * contained by) an existing feature. If there is no nesting, the feature is
   * added to the list and the method returns true. If nesting is found, the
   * feature is not added and the method returns false.
   * <p>
   * Contact features are added at the position of their first contact point
   * 
   * @param feature
   * @return
   */
  protected boolean addNonNestedFeature(SequenceFeature feature)
  {
    synchronized (nonNestedFeatures)
    {
      int insertPosition = binarySearchForAdd(nonNestedFeatures, feature);

      /*
       * fail if we detect feature enclosure - of the new feature by
       * the one preceding it, or of the next feature by the new one
       */
      if (insertPosition > 0)
      {
        if (encloses(nonNestedFeatures.get(insertPosition - 1), feature))
        {
          return false;
        }
      }
      if (insertPosition < nonNestedFeatures.size())
      {
        if (encloses(feature, nonNestedFeatures.get(insertPosition)))
        {
          return false;
        }
      }

      /*
       * checks passed - add or append the feature
       */
      if (insertPosition == nonNestedFeatures.size())
      {
        nonNestedFeatures.add(feature);
      }
      else
      {
        nonNestedFeatures.add(insertPosition, feature);
      }
      return true;
    }
  }

  /**
   * Answers true if range1 properly encloses range2, else false
   * 
   * @param range1
   * @param range2
   * @return
   */
  protected boolean encloses(ContiguousI range1, ContiguousI range2)
  {
    int begin1 = range1.getBegin();
    int begin2 = range2.getBegin();
    int end1 = range1.getEnd();
    int end2 = range2.getEnd();
    if (begin1 == begin2 && end1 > end2)
    {
      return true;
    }
    if (begin1 < begin2 && end1 >= end2)
    {
      return true;
    }
    return false;
  }

  /**
   * Answers the index of the first element in the given list which follows or
   * matches the given feature in the sort order. If no such element, answers
   * the length of the list.
   * 
   * @param list
   * @param feature
   * 
   * @return
   */
  protected int binarySearchForAdd(List<SequenceFeature> list, SequenceFeature feature)
  {
    // TODO binary search!
    int i = 0;
    while (i < list.size())
    {
      if (startOrdering.compare(nonNestedFeatures.get(i), feature) >= 0)
      {
        break;
      }
      i++;
    }
    return i;
  }

  /**
   * Add a contact feature to the lists that hold them ordered by start (first
   * contact) and by end (second contact) position, ensuring the lists remain
   * ordered, and returns true. If the contact feature lists already contain the
   * given feature (by test for equality), does not add it and returns false.
   * 
   * @param feature
   * @return
   */
  protected synchronized boolean addContactFeature(SequenceFeature feature)
  {
    if (contactFeatureStarts == null)
    {
      contactFeatureStarts = new ArrayList<SequenceFeature>();
    }
    if (contactFeatureEnds == null)
    {
      contactFeatureEnds = new ArrayList<SequenceFeature>();
    }

    // TODO binary search for insertion points!
    if (contactFeatureStarts.contains(feature))
    {
      return false;
    }

    contactFeatureStarts.add(feature);
    Collections.sort(contactFeatureStarts, startOrdering);
    contactFeatureEnds.add(feature);
    Collections.sort(contactFeatureEnds, endOrdering);

    return true;
  }

  /**
   * Returns a (possibly empty) list of entries whose range overlaps the given
   * range. The returned list is not ordered. Contact features are included if
   * either of the contact points lies within the range.
   * 
   * @param start
   *          start position of overlap range (inclusive)
   * @param end
   *          end position of overlap range (inclusive)
   * @return
   */
  public List<SequenceFeature> findOverlappingFeatures(long start, long end)
  {
    List<SequenceFeature> result = new ArrayList<SequenceFeature>();

    findNonNestedFeatures(start, end, result);

    findContactFeatures(start, end, result);

    if (nestedFeatures != null)
    {
      result.addAll(nestedFeatures.findOverlaps(start, end));
    }

    return result;
  }

  /**
   * Adds contact features to the result list where either the second or the
   * first contact position lies within the target range.
   * 
   * @param from
   * @param to
   * @param result
   */
  protected void findContactFeatures(long from, long to,
          List<SequenceFeature> result)
  {
    if (contactFeatureStarts != null)
    {
      findContactStartFeatures(from, to, result);
    }
    if (contactFeatureEnds != null)
    {
      findContactEndFeatures(from, to, result);
    }
  }

  /**
   * @param from
   * @param to
   * @param result
   */
  protected void findContactEndFeatures(long from, long to,
          List<SequenceFeature> result)
  {
    // TODO binary search for startPosition
    for (int startPosition = 0; startPosition < contactFeatureEnds.size(); startPosition++)
    {
      SequenceFeature sf = contactFeatureEnds.get(startPosition);
      if (!sf.isContactFeature())
      {
        System.err.println("Error! non-contact feature type "
                + sf.getType() + " in contact features list");
        continue;
      }
      int begin = sf.getBegin();
      if (begin >= from && begin <= to)
      {
        /*
         * this feature's first contact position lies in the search range
         * so we don't include it in results a second time
         */
        continue;
      }
      int end = sf.getEnd();
      if (end >= from && end <= to)
      {
        result.add(sf);
      }
    }
  }

  /**
   * Returns the index of the first contact feature found whose end (second
   * contact position) is not before the given start position. If no such
   * feature is found, returns the length of the contact features list.
   * 
   * @param start
   * @return
   */
  protected int contactsBinarySearch(long start)
  {
    // TODO binary search!!
    int i = 0;
    while (i < contactFeatureEnds.size())
    {
      if (contactFeatureEnds.get(i).getEnd() >= start)
      {
        break;
      }
      i++;
    }

    return i;
  }

  /**
   * Adds features to the result list that are at a single position which lies
   * within the target range. Non-positional features (start=end=0) and contact
   * features are excluded.
   * 
   * @param from
   * @param to
   * @param result
   */
  protected void findNonNestedFeatures(long from, long to,
          List<SequenceFeature> result)
  {
    int startIndex = binarySearch(nonNestedFeatures, from);
    findNonNestedFeatures(startIndex, from, to, result);
  }

  /**
   * Scans the list of non-nested features, starting from startIndex, to find
   * those that overlap the from-to range, and adds them to the result list.
   * Returns the index of the first feature whose start position is after the
   * target range (or the length of the whole list if none such feature exists).
   * 
   * @param startIndex
   * @param from
   * @param to
   * @param result
   * @return
   */
  protected int findNonNestedFeatures(final int startIndex, long from,
          long to,
          List<SequenceFeature> result)
  {
    int i = startIndex;
    while (i < nonNestedFeatures.size())
    {
      SequenceFeature sf = nonNestedFeatures.get(i);
      if (sf.getBegin() > to)
      {
        break;
      }
      int start = sf.getBegin();
      int end = sf.getEnd();
      if (start <= to && end >= from)
      {
        result.add(sf);
      }
      i++;
    }
    return i;
  }

  /**
   * Performs a binary search of the (sorted) list to find the index of the
   * first entry whose end position is not less than the target position (i.e.
   * skip all features that properly precede the given position)
   * 
   * @param features
   * @param target
   * @return
   */
  protected int binarySearch(List<SequenceFeature> features, long target)
  {
    int width = features.size() / 2;
    int lastpos = width;
    while (width > 0)
    {
      int end = features.get(lastpos).getEnd();
      width = width / 2;
      if (end > target)
      {
        lastpos -= width;
      }
      else
      {
        lastpos += width;
      }
    }
    // todo correct binary search
    return lastpos > 1 ? lastpos - 2 : 0;
    // return lastpos;
  }

  /**
   * Adds contact features whose start position lies in the from-to range to the
   * result list
   * 
   * @param from
   * @param to
   * @param result
   */
  protected void findContactStartFeatures(long from, long to,
          List<SequenceFeature> result)
  {
    // TODO binary search for startPosition
    for (int startPosition = 0; startPosition < contactFeatureStarts.size(); startPosition++)
    {
      SequenceFeature sf = contactFeatureStarts.get(startPosition);
      if (!sf.isContactFeature())
      {
        System.err.println("Error! non-contact feature type "
                + sf.getType() + " in contact features list");
        continue;
      }
      int begin = sf.getBegin();
      if (begin >= from && begin <= to)
      {
        result.add(sf);
      }
    }
  }

  /**
   * Answers a list of all positional features stored, in no guaranteed order
   * 
   * @return
   */
  public List<SequenceFeature> getPositionalFeatures()
  {
    /*
     * add non-nested features (may be all features for many cases)
     */
    List<SequenceFeature> result = new ArrayList<SequenceFeature>();
    result.addAll(nonNestedFeatures);

    /*
     * add any contact features - from the list by start position
     */
    if (contactFeatureStarts != null)
    {
      result.addAll(contactFeatureStarts);
    }

    /*
     * add any nested features
     */
    if (nestedFeatures != null)
    {
      result.addAll(nestedFeatures.getEntries());
    }

    return result;
  }

  /**
   * Answers a list of all contact features. If there are none, returns an
   * immutable empty list.
   * 
   * @return
   */
  public List<SequenceFeature> getContactFeatures()
  {
    if (contactFeatureStarts == null)
    {
      return Collections.emptyList();
    }
    return new ArrayList<SequenceFeature>(contactFeatureStarts);
  }

  /**
   * Answers a list of all non-positional features. If there are none, returns
   * an immutable empty list.
   * 
   * @return
   */
  public List<SequenceFeature> getNonPositionalFeatures()
  {
    if (nonPositionalFeatures == null)
    {
      return Collections.emptyList();
    }
    return new ArrayList<SequenceFeature>(nonPositionalFeatures);
  }

  /**
   * Deletes the given feature from the store, returning true if it was found
   * (and deleted), else false. This method makes no assumption that the feature
   * is in the 'expected' place in the store, in case it has been modified since
   * it was added.
   * 
   * @param sf
   */
  public synchronized boolean delete(SequenceFeature sf)
  {
    /*
     * try the non-nested positional features first
     */
    boolean removed = nonNestedFeatures.remove(sf);

    /*
     * if not found, try contact positions (and if found, delete
     * from both lists of contact positions)
     */
    if (!removed && contactFeatureStarts != null)
    {
      removed = contactFeatureStarts.remove(sf);
      if (removed)
      {
        contactFeatureEnds.remove(sf);
      }
    }

    boolean removedNonPositional = false;

    /*
     * if not found, try non-positional features
     */
    if (!removed && nonPositionalFeatures != null)
    {
      removedNonPositional = nonPositionalFeatures.remove(sf);
      removed = removedNonPositional;
    }

    /*
     * if not found, try nested features
     */
    if (!removed && nestedFeatures != null)
    {
      removed = nestedFeatures.delete(sf);
    }

    if (removed)
    {
      rebuildFeatureGroups(sf.getFeatureGroup(), removedNonPositional);
    }

    return removed;
  }

  /**
   * Check whether the given feature group is still represented, in either
   * positional or non-positional features, and if not, remove it from the set
   * of feature groups
   * 
   * @param featureGroup
   * @param nonPositional
   */
  protected void rebuildFeatureGroups(String featureGroup,
          boolean nonPositional)
  {
    if (nonPositional && nonPositionalFeatures != null)
    {
      boolean found = false;
      for (SequenceFeature sf : nonPositionalFeatures)
      {
        String group = sf.getFeatureGroup();
        if (featureGroup == group
                || (featureGroup != null && featureGroup.equals(group)))
        {
          found = true;
          break;
        }
      }
      if (!found)
      {
        nonPositionalFeatureGroups.remove(featureGroup);
      }
    }
    else if (!findFeatureGroup(featureGroup))
    {
      positionalFeatureGroups.remove(featureGroup);
    }
  }

  /**
   * Scans all positional features to check whether the given feature group is
   * found, and returns true if found, else false
   * 
   * @param featureGroup
   * @return
   */
  protected boolean findFeatureGroup(String featureGroup)
  {
    for (SequenceFeature sf : getPositionalFeatures())
    {
      String group = sf.getFeatureGroup();
      if (group == featureGroup
              || (group != null && group.equals(featureGroup)))
      {
        return true;
      }
    }
    return false;
  }

  /**
   * Answers true if this store has no features, else false
   * 
   * @return
   */
  public boolean isEmpty()
  {
    boolean hasFeatures = !nonNestedFeatures.isEmpty()
            || (contactFeatureStarts != null && !contactFeatureStarts
                    .isEmpty())
            || (nonPositionalFeatures != null && !nonPositionalFeatures
                    .isEmpty())
            || (nestedFeatures != null && nestedFeatures.size() > 0);

    return !hasFeatures;
  }

  /**
   * Answers the set of distinct feature groups stored, possibly including null,
   * as an unmodifiable view of the set. The parameter determines whether the
   * groups for positional or for non-positional features are returned.
   * 
   * @param positionalFeatures
   * @return
   */
  public Set<String> getFeatureGroups(boolean positionalFeatures)
  {
    if (positionalFeatures)
    {
      return Collections.unmodifiableSet(positionalFeatureGroups);
    }
    else
    {
      return nonPositionalFeatureGroups == null ? Collections
              .<String> emptySet() : Collections
              .unmodifiableSet(nonPositionalFeatureGroups);
    }
  }
}
